from __future__ import print_function, division
import numpy as np
import scipy.interpolate


def detrend_huber_spline(time, flux, mask, knot_distance):
    """Robust B-Spline regression with scikit-learn"""
    try:
        from sklearn.base import TransformerMixin
        from sklearn.pipeline import make_pipeline
        from sklearn.linear_model import HuberRegressor
    except:
        raise ImportError('Could not import sklearn')

    class BSplineFeatures(TransformerMixin):
        def __init__(self, knots):
            self.bsplines = self.get_bspline_basis(knots)

        def fit(self, X, y=None):
            return self

        def transform(self, X):
            features = np.zeros((len(X), len(self.bsplines)))
            for ispline, spline in enumerate(self.bsplines):
                features[:, ispline:ispline+1] = scipy.interpolate.splev(X, spline)
            return features

        def get_bspline_basis(self, knots):
            knots, coeffs, degree = scipy.interpolate.splrep(knots, np.zeros(len(knots)))
            bsplines = []
            for ispline in range(len(knots)):
                coeffs = [1 if ispl == ispline else 0 for ispl in range(len(coeffs))]
                bsplines.append((knots, coeffs, degree))
            return bsplines

    masked_flux = flux[mask==1]
    masked_time = time[mask==1]

    if len(masked_time) == 0:
        return np.full(len(time), np.nan)
    else:
        duration = np.max(masked_time) - np.min(masked_time)
        no_knots = int(duration / knot_distance)
        knots = np.linspace(np.min(masked_time), np.max(masked_time), no_knots)

    if len(knots) < 4:
        return np.full(len(time), np.nan)
    else:
        pipeline = make_pipeline(BSplineFeatures(knots), HuberRegressor())
        trend = pipeline.fit(masked_time[:, np.newaxis], masked_flux)
        return trend.predict(time[:, None])



from __future__ import print_function, division
import numpy as np
from numba import jit


@jit(fastmath=True, nopython=True, cache=True)
def lowess(x, y, mask, window_length, maxiter=30, ftol=1e-6):

    def calc_x_weights(x, xi, y_weights):
        radius = max(xi - x[0], x[-2] - xi)
        x_weights = (1 - (np.abs(x - xi) / radius) ** 3) ** 3  # tricube
        if np.any(y_weights):
            x_weights = x_weights * y_weights
        return x_weights / np.nansum(x_weights)

    def calc_y_weights(y, mask, trend):
        diff = np.abs(y - trend)
        diff /= 6 * np.nanmedian(diff) + 1e-100  # avoid division by zero if median == 0
        diff[diff > 1] = 1
        y_weights = (1 - diff**2)**2
        y_weights[mask == 0] = 0
        return y_weights  # bisquare

    def calc_y_fit(x, y, xi, weights):  # "projection vector" as in statsmodels
        w1 = np.sum(weights * x)  #
        w2 = np.sum(weights * (x - w1)**2)
        if w2 == 0:
            return np.nan
        else:
            return np.sum(weights * ((1 + (xi - w1) * (x - w1) / w2) * y))

    points = len(x)
    y_fit_previous = np.ones(points)
    y_weights = np.zeros(points)
    for iter in range(maxiter):
        left = 0
        if (np.max(x) - np.min(x)) == 0:
            return np.full(len(x), np.nan)
        right = int((window_length / (np.max(x) - np.min(x))) * points)
        trend = np.zeros(points)
        for i in range(points):
            while right < points and (x[i] > (x[left] + x[right]) / 2):
                left += 1  # move the window forward
                right += 1
            x_weights = calc_x_weights(x[left:right], x[i], y_weights[left:right])
            trend[i] = calc_y_fit(x[left:right], y[left:right], x[i], x_weights)
        y_weights = calc_y_weights(y, mask, trend)
        if np.max(np.abs(y_fit_previous - trend)) < ftol:  # Check convergence
            break
        y_fit_previous = trend
    return trend



from __future__ import print_function, division
import numpy as np
import wotan.constants as constants
from wotan.helpers import cleaned_array


def pspline(time, flux, edge_cutoff, max_splines, return_nsplines, verbose):
    try:
        from pygam import LinearGAM, s
    except:
        raise ImportError('Could not import pygam')

    newflux = flux.copy()
    newtime = time.copy()
    detrended_flux = flux.copy()/np.nanmedian(newflux)

    for i in range(constants.PSPLINES_MAXITER):
        mask_outliers = np.ma.where(
            np.abs(1-detrended_flux) < constants.PSPLINES_STDEV_CUT*np.std(detrended_flux))
        if len(mask_outliers[0]) != 0:  # Do not cut if zero points remain
            newtime, newflux = cleaned_array(newtime[mask_outliers], newflux[mask_outliers])
        gam = LinearGAM(s(0, n_splines=max_splines))
        search_gam = gam.gridsearch(newtime[:, np.newaxis], newflux, progress=False)
        trend = search_gam.predict(newtime)
        detrended_flux = newflux / trend
        stdev = np.std(detrended_flux)
        mask_outliers = np.ma.where(
            np.abs(1-detrended_flux) > constants.PSPLINES_STDEV_CUT*np.std(detrended_flux))
        if verbose:
            print('Iteration:', i + 1, 'Rejected outliers:', len(mask_outliers[0]))
            # Check convergence
            if len(mask_outliers[0]) == 0:
                print('Converged.')
                break

    # Final iteration, applied to unclipped time series (interpolated over clipped values)
    mask_outliers = np.ma.where(np.abs(1-detrended_flux) < constants.PSPLINES_STDEV_CUT*stdev)
    if len(mask_outliers[0]) != 0:  # Do not cut if zero points remain
        newtime, newflux = cleaned_array(newtime[mask_outliers], newflux[mask_outliers])
    gam = LinearGAM(s(0, n_splines=max_splines))
    search_gam = gam.gridsearch(newtime[:, np.newaxis], newflux, progress=False)
    trend = search_gam.predict(time)

    # Cut off edges
    if edge_cutoff > 0:
        low_index = np.argmax(time > (min(time)+edge_cutoff))
        hi_index = np.argmax(time > (max(time)-edge_cutoff))
        trend[:low_index] = np.nan
        trend[hi_index:] = np.nan

    nsplines = np.ceil(gam.statistics_['edof'])
    return trend, nsplines



from __future__ import division
import numpy as np


class RobustNorm(object):
    def __call__(self, z):
        return self.rho(z)


class HuberT(RobustNorm):
    def __init__(self, t=1.5):
        self.t = t

    def _subset(self, z):
        return np.less_equal(np.fabs(z), self.t)

    def weights(self, z):
        test = self._subset(z)
        absz = np.fabs(z)
        absz[test] = 1.0
        #absz[absz==0] = 1e-10
        return test + (1 - test) * self.t / absz


class RamsayE(RobustNorm):
    def __init__(self, a=0.3):
        self.a = a

    def weights(self, z):
        return np.exp(-self.a * np.fabs(z))


class Hampel(RobustNorm):
    def __init__(self, a=1.7, b=3.4, c=8.5):
        self.a = a
        self.b = b
        self.c = c

    def _subset(self, z):
        z = np.fabs(z)
        t1 = np.less_equal(z, self.a)
        t2 = np.less_equal(z, self.b) * np.greater(z, self.a)
        t3 = np.less_equal(z, self.c) * np.greater(z, self.b)
        return t1, t2, t3

    def weights(self, z):
        a = self.a
        b = self.b
        c = self.c
        t1, t2, t3 = self._subset(z)
        fabs = np.fabs(z)
        fabs[fabs==0] = 1e-100
        v = (
            t1
            + t2 * a / fabs
            + t3
            * a
            * (c - fabs)
            / (fabs * (c - b))
        )
        v[np.where(np.isnan(v))] = 1
        return v





def estimate_location(data, norm, maxiter=30, tol=1e-6):

    if norm == "huber":
        norm = HuberT()
    elif norm == "ramsay":
        norm = RamsayE()
    elif norm == "hampel":
        norm = Hampel()
    location = np.median(data)  # Initial guess
    #print(location)

    for iter in range(maxiter):
        W = norm.weights((data - location))
        #print(W)
        new_location = np.sum(W * data) / np.sum(W)
        #print(iter, new_location)
        if np.alltrue(np.less(np.fabs(location - new_location), tol)):
            return new_location
        else:
            location = new_location
    # Did not converge
    print(norm, 'Not converged')
    return np.median(data)



def ramsay(data, maxiter=30, tol=1e-6):
    
    a = 0.3
    location = np.median(data)  # Initial guess
    for iter in range(maxiter):
        z = data - location
        W = np.exp(-a * np.fabs(z))
        new_location = np.sum(W * data) / np.sum(W)
        #print(iter, new_location)
        if np.alltrue(np.less(np.fabs(location - new_location), tol)):
            return new_location
        else:
            location = new_location
    return np.median(data)



def huber(data, maxiter=30, tol=1e-6):
    
    t = 1.5
    location = np.median(data)  # Initial guess
    for iter in range(maxiter):
        z = data - location

        test = np.less_equal(np.fabs(z), t)
        absz = np.fabs(z)
        absz[test] = 1
        absz[absz==0] = 1e-100
        W = test + (1 - test) * t / absz

        new_location = np.sum(W * data) / np.sum(W)
        #print(iter, new_location)
        if np.alltrue(np.less(np.fabs(location - new_location), tol)):
            return new_location
        else:
            location = new_location
    return np.median(data)


def hampel(data, maxiter=30, tol=1e-6):
    
    a=1.7
    b=3.4
    c=8.5

    location = np.median(data)  # Initial guess
    for iter in range(maxiter):
        z = data - location

        fabs = np.fabs(z)
        fabs[fabs==0] = 1e-100
        W = (
            np.less_equal(z, a) + np.less_equal(z, b) * np.greater(z, a) * a / fabs
            + np.less_equal(z, c) * np.greater(z, b) * a * (c - fabs) / (fabs * (c - b))
            )
        W[np.where(np.isnan(W))] = 1
        

        new_location = np.sum(W * data) / np.sum(W)
        #print(iter, new_location)
        if np.alltrue(np.less(np.fabs(location - new_location), tol)):
            return new_location
        else:
            location = new_location
    return np.median(data)




data = [0, 1, 2, 3, 11, 11.5]

print('ramsay old')
print(estimate_location(data, norm="ramsay"))
print('ramsay new')
print(ramsay(data))

print('huber old')
print(estimate_location(data, norm="huber"))
print('huber new')
print(huber(data))

print('hampel old')
print(estimate_location(data, norm="hampel"))
print('hampel new')
print(hampel(data))



from __future__ import division
import numpy as np


def regression(time, flux, method, window_length, cval):
    """Regressions with scikit-learn"""

    try:
        from sklearn.pipeline import make_pipeline
        from sklearn.linear_model import Ridge, Lasso, ElasticNet
        from sklearn.base import BaseEstimator, TransformerMixin
    except:
        raise ImportError('Could not import sklearn')

    class GF(BaseEstimator, TransformerMixin):
        """Uniformly spaced Gaussian features for one-dimensional input"""
    
        def __init__(self, N, width_factor=2.0):
            self.N = N
            self.width_factor = width_factor
        
        @staticmethod
        def _gauss_basis(x, y, width, axis=None):
            arg = (x - y) / width
            return np.exp(-0.5 * np.sum(arg ** 2, axis))
            
        def fit(self, X, y=None):
            # create N centers spread along the data range
            self.centers_ = np.linspace(X.min(), X.max(), self.N)
            self.width_ = self.width_factor * (self.centers_[1] - self.centers_[0])
            return self
            
        def transform(self, X):
            return self._gauss_basis(X[:, :, np.newaxis], self.centers_,
                                     self.width_, axis=1)

    offset = 1e-6  # lasso and elasticnet are less sensitive than ridge regression
    tol = 1e-6
    max_iter = 1e5
    if method == 'ridge':
        regression_method = Ridge(alpha=cval, max_iter=max_iter, tol=tol)
    elif method == 'lasso':
        regression_method = Lasso(alpha=cval * offset, max_iter=max_iter, tol=tol)
    elif method == 'elasticnet':
        regression_method = ElasticNet(alpha=cval * offset, max_iter=max_iter, tol=tol)

    duration = np.max(time) - np.min(time)
    no_knots = int(duration / window_length)
    X = time[:, np.newaxis]
    trend = make_pipeline(GF(no_knots), regression_method).fit(X, flux).predict(X)

    return trend



from __future__ import print_function, division
from numba import jit
import numpy as np
import wotan.constants as constants


@jit(fastmath=True, nopython=True, cache=True)
def matrix_gen(t, degree):
    dur = 2 * (np.max(t) - np.min(t))
    rows = len(t)
    cols = 2 * (degree + 1)
    matrix = np.ones(shape=(rows, cols))
    for x in range(rows):
        for y in range(1, int(cols / 2)):
            val = (2 * np.pi * t[x] * y) / dur
            matrix[x, y * 2] = np.sin(val)
            matrix[x, y * 2 + 1] = np.cos(val)
        matrix[x, 1] = t[x]
    return matrix


def detrend_cosine(t, y, window_length, robust, mask):
    degree = (int((max(t) - min(t)) / window_length))
    #if not robust:
    #    constants.PSPLINES_MAXITER = 0
        #matrix = matrix_gen(t, degree)
        #trend = np.matmul(matrix, np.linalg.lstsq(matrix, y, rcond=-1)[0])

    # robust version: sigma-clip flux from trend and iterate until convergence
    no_clip_previous = np.inf
    if not robust:
        converged = True
    else:
        converged = False
    for i in range(constants.PSPLINES_MAXITER):
        matrix = matrix_gen(t, degree)
        # Add weights in order to weight down the masked values
        # Solution from https://stackoverflow.com/questions/27128688/how-to-use-least-squares-with-weight-matrix
        Aw = matrix * mask[:,np.newaxis]  # if real weights: sqrt
        Bw = y * mask  # if real weights: sqrt
        trend = np.matmul(matrix, np.linalg.lstsq(Aw, Bw, rcond=None)[0])
        detrended_flux = y / trend
        mask_outliers = np.ma.where(
            1-detrended_flux > constants.PSPLINES_STDEV_CUT*np.std(detrended_flux))
        mask[mask_outliers] = 1e-10
        if no_clip_previous == len(mask_outliers[0]):
            converged = True
        no_clip_previous = len(mask_outliers[0])
        if converged:
            print('Converged.')
            break
        else:
            print('Iteration:', i + 1, 'Rejected outliers (total):', len(mask_outliers[0]))
    return trend


def detrend_cofiam(t, y, window_length):
    degree = (int((max(t) - min(t)) / window_length))
    dw_previous = np.inf
    dw_mask = np.array([True] * len(t))
    for k_m in range(1, degree + 1):
        matrix = matrix_gen(t, degree)
        trend = np.matmul(matrix, np.linalg.lstsq(matrix, y, rcond=-1)[0])

        # Durbin-Watson autocorrelation statistics
        dw_y = y[dw_mask] / trend[dw_mask] - 1
        dw = np.abs(np.sum((dw_y[1:] - dw_y[:-1]) ** 2) / (np.sum(dw_y ** 2)) - 2)

        # If Durbin-Watson *increased* this round: Previous was the best
        if dw > dw_previous:
            return trend
        dw_previous = dw
    matrix = matrix_gen(t, degree)
    trend = np.matmul(matrix, np.linalg.lstsq(matrix, y, rcond=-1)[0])
    return trend



from __future__ import print_function, division
import numpy as np
import wotan.constants as constants
from wotan.helpers import cleaned_array
from scipy.interpolate import LSQUnivariateSpline


def iter_spline(time, flux, mask, window_length):
    masked_flux = flux[mask==1]
    masked_time = time[mask==1]
    no_knots = int((max(time) - min(time)) / window_length)
    newflux = masked_flux.copy()
    newtime = masked_time.copy()
    newtime, newflux = cleaned_array(newtime, newflux)
    detrended_flux = masked_flux.copy()
    for i in range(constants.PSPLINES_MAXITER):
        outliers = 1 - detrended_flux < constants.PSPLINES_STDEV_CUT * np.nanstd(detrended_flux)
        mask_outliers = np.ma.where(outliers)
        newtime, newflux = cleaned_array(newtime[mask_outliers], newflux[mask_outliers])
        # knots must not be at the edges, so we take them as [1:-1]
        if len(newtime) < 5:
            return np.full(len(time), np.nan)
        knots = np.linspace(min(newtime), max(newtime), no_knots)[1:-1]
        try:
            trend = LSQUnivariateSpline(newtime, newflux, knots)
        except:
            return np.full(len(time), np.nan)
        trend_segment = trend(newtime)
        detrended_flux = newflux / trend_segment
        outliers = 1 - detrended_flux > constants.PSPLINES_STDEV_CUT * np.nanstd(detrended_flux)
        mask_outliers = np.ma.where(mask_outliers)
        if len(mask_outliers[0]) == 0:
            break
    return trend(time)  # Final iteration: apply trend to clipped and masked values
